#
# Copyright 2002 Roland Stigge
#
# This file is part of ImageWorkshop
#
# ImageWorkshop is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# ImageWorkshop is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ImageWorkshop; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

# Tkinter
import Tkinter
import os

class ProgressBar(Tkinter.Frame):
    """
    A progress bar

    originally from: "Python and Tkinter programming" by John Grayson
    """
    def __init__(self, master=None, orientation="horizontal",
                 min=0, max=100, width=100, height=18,
                 doLabel=1, appearance="sunken",
                 fillColor="blue", background="gray",
                 labelColor="yellow", labelFont="Courier",
                 labelText="", labelFormat="%d%%",
                 value=50, bd=2):
        # preserve various values
        self.master=master
        self.orientation=orientation
        self.min=min
        self.max=max
        self.width=width
        self.height=height
        self.doLabel=doLabel
        self.fillColor=fillColor
        self.labelFont= labelFont
        self.labelColor=labelColor
        self.background=background
        self.labelText=labelText
        self.labelFormat=labelFormat
        self.value=value
        Tkinter.Frame.__init__(self, master, relief=appearance, bd=bd)
        self.canvas=Tkinter.Canvas(self, height=height, width=width,
                                   bd=0, highlightthickness=0,
                                   background=background)
        self.scale=self.canvas.create_rectangle(0, 0, width, height,
                                                fill=fillColor)
        self.label=self.canvas.create_text(self.canvas.winfo_reqwidth() / 2,
                                           height / 2, text=labelText,
                                           anchor="c", fill=labelColor,
                                           font=self.labelFont)
        self.updateFrame()
        self.canvas.pack(side='top', fill='x', expand='no')

    def updateProgress(self, newValue, newMax=None):
        "update method, to be called by the application"
        if newMax:
            self.max = newMax
        self.value = newValue
        self.updateFrame()

    def updateFrame(self):
        # Trim the values to be between min and max
        value=self.value
        if value > self.max:
            value = self.max
        if value < self.min:
            value = self.min
        # Adjust the rectangle
        if self.orientation == "horizontal":
            self.canvas.coords(self.scale, 0, 0,
              float(value) / self.max * self.width, self.height)
        else:
            self.canvas.coords(self.scale, 0,
                               self.height - (float(value) / 
                                              self.max*self.height),
                               self.width, self.height)
        # Now update the colors
        self.canvas.itemconfig(self.scale, fill=self.fillColor)
        self.canvas.itemconfig(self.label, fill=self.labelColor)
        # And update the label
        if self.doLabel:
            if value:
                if value >= 0:
                    pvalue = int((float(value) / float(self.max)) * 
                                   100.0)
                else:
                    pvalue = 0
                self.canvas.itemconfig(self.label, text=self.labelFormat
                                         % pvalue)
            else:
                self.canvas.itemconfig(self.label, text='')
        else:
            self.canvas.itemconfig(self.label, text=self.labelFormat %
                                   self.labelText)
        self.canvas.update_idletasks()
        
class Dialog(Tkinter.Toplevel):
    """
    A freely configurable dialog
    override:
      body:      fill the dialog window, return widget with initial focus
      buttonbox: buttons at bottom of window
      validate:  test dialog contents for correctness
      apply:     if everything is correct, apply contents

    originally from: "An Introduction to Tkinter" by Fredrik Lundh
    """
    
    def __init__(self, parent, title = None):
        Tkinter.Toplevel.__init__(self, parent)
        self.transient(parent) # don't show up in window list!

        if title:              # put name into title bar
            self.title(title)

        self.parent = parent

        self.result = None

        body = Tkinter.Frame(self)
        self.initial_focus = self.body(body)
        body.pack(padx=5, pady=5)

        self.buttonbox()

        self.grab_set()

        if not self.initial_focus:
            self.initial_focus = self

        self.protocol("WM_DELETE_WINDOW", self.cancel)

        # set window position
        #self.geometry("+%d+%d" % (parent.winfo_rootx()+50,
        #                          parent.winfo_rooty()+50))

        self.initial_focus.focus_set()

        self.wait_window(self)

    #
    # construction hooks
    #

    def body(self, master):
        """
        create dialog body
        master will be a Tkinter.Frame
        return widget that should have initial focus
        this method should be overridden
        """
        pass

    def buttonbox(self):
        """
        add standard button box. override if you don't want the
        standard buttons
        """
              
        box = Tkinter.Frame(self)

        w = Tkinter.Button(box, text="OK", width=10, command=self.ok, default=Tkinter.ACTIVE)
        w.pack(side=Tkinter.LEFT, padx=5, pady=5)
        w = Tkinter.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tkinter.LEFT, padx=5, pady=5)
        
        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)
        
        box.pack()

    #
    # standard button semantics
    #

    def ok(self, event=None):
        "ok has been pressed"
        if not self.validate():
            self.initial_focus.focus_set() # put focus back
            return

        self.withdraw()
        self.update_idletasks()

        self.apply()

        self.cancel()

    def cancel(self, event=None):
        "cancel has been pressed: put focus back to the parent window"
        self.parent.focus_set()
        self.destroy()

    #
    # command hooks
    #

    def validate(self):
        """
        validate entries for correctness
        return 1 if ok
        (override this method if correctness matters!
         maybe apply is then not needed.)
        """
        return 1

    def apply(self):
        """
        carry out actions associated with the OK button
        (override this method if it should do something!)
        """
        pass

class BigInfo(Dialog):
    """
    not a real dialog, just a freely configurable info window
    override: body
    """

    def buttonbox(self):
        "An OK button is enough!"
        
        w = Tkinter.Button(self, text="OK", width=10, command=self.ok, default=Tkinter.ACTIVE)
        w.pack(padx=5, pady=5)

        self.bind("<Return>", self.ok)
        
def getmenuname(s):
    """
    input: Menuentry specification string, e.g. "_Help"
    output: Menuentry display string, e.g. "Help"
            and position of first occurence of "_"
            -> tuple (string, int)
    if no "_" is specified, int value (second in tuple) is 0
    """
    try:
        pos = s.index("_")
    except ValueError:
        return (s, -1)
    return (s[:pos] + s[pos + 1:], pos)

def fillmenu(menu, structure, master):
    """
    fill a given menu (Tkinter.Menu) with the given structure:
    a list of items; recursive: description of possible lines:
       (1) "cascade", "<namespec>", "<object name in dict>", structure
       (2) "separator"
       (3) "command", "<namespec>", command, "<accelerator>"
       (4) "radiobutton", "<namespec>", command, "<accelerator>", variable,
           value
       (5) "checkbutton", "<namespec>", command, "<accelerator>", variable
    master is the object where to place the cascade (Menu) objects into
      __dict__ and where the shortcut bindings go

    structure is again a list of items of a cascaded menu
    accelerator is an X shortcut binding, e.g. "<Alt-x>" or "+"
    """
    for i in structure:
        keywords = {} # arguments for menu.add()
        if i[0] != "separator":
            (name, pos) = getmenuname(i[1])
            keywords["label"] = name
            if pos >= 0:
                keywords["underline"] = pos

        if (i[0] in ["command", "radiobutton", "checkbutton"]) and i[3]:
            # accelerator specified
            if i[3][0] == "<" and i[3][-1] == ">":
                disp = i[3][1:-1]
            else:
                disp = i[3]
            keywords["accelerator"] = disp
            master.bind(i[3], i[2])

        if i[0] == "cascade":
            newmenu = Tkinter.Menu(menu)
            fillmenu(newmenu, i[3], master)
            keywords["menu"] = newmenu
            master.__dict__[i[2]] = newmenu
            
        elif i[0] == "command":
            keywords["command"] = i[2]
            
        elif i[0] == "radiobutton":
            keywords["command"] = i[2]
            keywords["variable"] = i[4]
            keywords["value"] = i[5]
            
        elif i[0] == "checkbutton":
            keywords["command"] = i[2]
            keywords["variable"] = i[4]

        apply(menu.add, [i[0]], keywords)
