#
# Copyright 2002 Roland Stigge
#
# This file is part of ImageWorkshop
#
# ImageWorkshop is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# ImageWorkshop is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ImageWorkshop; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

# system modules
import math

# Tkinter
import Tkinter

# Python Imaging Library (PIL)
import Image

# Numerical Python
import Numeric
import LinearAlgebra

# own modules
import GUI
import ImageGUI         # Image handling
import NumericImage

#
# classes for global operations
#

class FourierImageWindow(ImageGUI.ColorImageWindow):
    "represents a Fourier transformed image"

    def __init__(self, master, way = 1): # way: 1 = normal, 0 = inverse
        self.way = way
        masterimage = master.getnumericimage() # may be complex
        DFT1 = NumericImage.fouriermatrix(master.height())
        DFT2 = NumericImage.fouriermatrix(master.width())

        self.fourier = []
        for i in masterimage:
            if way:
                self.fourier.append(Numeric.matrixmultiply(
                    Numeric.matrixmultiply(DFT1, i._get_M()),
                    LinearAlgebra.inverse(DFT2)))
            else:
                self.fourier.append(Numeric.matrixmultiply(
                    Numeric.matrixmultiply(LinearAlgebra.inverse(DFT1),
                                           i._get_M()), DFT2))

        self.centermode = Tkinter.IntVar()
        if way:
            self.centermode.set(1) # centered mode
            titleaddon = "DFT"
        else: titleaddon = "inverseDFT"
        self.dftviewmode = Tkinter.IntVar()
        self.dftviewmode.set(0)
        # 0 = abs, 1 = abs+log, 2 = radian, 3 = real, 4 = imag

        ImageGUI.ColorImageWindow.__init__(self, master,
                                           master.name + " " + titleaddon)

        if way:
            addon_structure = [
                ("separator",),
                ("checkbutton", "Centere_d", self.changeviewmode, "",
                 self.centermode),
                ("separator",),
                ("radiobutton", "_Absolute", self.changeviewmode, "", self.dftviewmode, 0),
                ("radiobutton", "Absolute, _logarithmic", self.changeviewmode, "", self.dftviewmode, 1),
                ("radiobutton", "Radia_n", self.changeviewmode, "", self.dftviewmode, 2),
                ("radiobutton", "R_eal", self.changeviewmode, "", self.dftviewmode, 3),
                ("radiobutton", "I_maginary", self.changeviewmode, "", self.dftviewmode, 4)
                ]
            GUI.fillmenu(self.viewmenu, addon_structure, self)

            addon_structure = [
                ("separator",),
                ("command", "_Inverse Fourier Transform",
                 self.inverse_fourier, "")]
            GUI.fillmenu(self.globalmenu, addon_structure, self)

    def inverse_fourier(self, event = None):
        "Inverse Fourier Transform"
        FourierImageWindow(self, way = 0) # inverse transform

    def getrawmode(self):
        "implements for ColorImageWindow"
        if len(self.fourier) == 1: return "L"
        else: return "RGB"

    def getrawview(self):
        "implements for ColorImageWindow"
        result = []
        for i in self.fourier:
            mode = self.dftviewmode.get()
            if mode in [0, 1]: # absolute
                channel = Numeric.absolute(i)
                if mode == 1: channel = Numeric.log(channel + 1)
            elif mode == 2: # radian
                channel = Numeric.zeros((i.shape), Numeric.Float)
                for r in range(channel.shape[0]):
                    for c in range(channel.shape[1]):
                        value = i[r, c]
                        if value == 0:
                            channel[r, c] = 0
                        elif value.imag >= 0:
                            channel[r, c] = +math.acos(value.real / abs(value))
                        else:
                            channel[r, c] = -math.acos(value.real / abs(value))
                channel = (channel / math.pi + 1) * 255 / 2
            elif mode == 3: # real
                channel = i.real
            else: # 4: imag
                channel = i.imag

            result.append(channel)

        if mode != 2 and self.way: # normalization
            maximum = max(max(max(result)))
            for i in range(len(result)):
                if maximum:
                    result[i] = result[i] * 255 / maximum
                else: # avoid division by zero
                    result[i] = result[i] * 0

        for i in range(len(result)): # -> to Image
            result[i] = NumericImage.NumImage(result[i])
            if self.centermode.get():
                result[i] = result[i].cornerstocenter()
            
        if len(self.fourier) == 1: # "L"
            return result[0].toImage()
        else: # "RGB"
            return Image.merge("RGB", [result[0].toImage(),
                                       result[1].toImage(),
                                       result[2].toImage()])

    def getrawnumericimage(self):
        "implements for ColorImageWindow"
        if len(self.fourier) == 1:
            return [NumericImage.NumImage(self.fourier[0])]
        else:
            return [NumericImage.NumImage(self.fourier[0]),
                    NumericImage.NumImage(self.fourier[1]),
                    NumericImage.NumImage(self.fourier[2])]
        
    def width(self):
        "implements for ColorImageWindow"
        return self.fourier[0].shape[1]
        
    def height(self):
        "implements for ColorImageWindow"
        return self.fourier[0].shape[0]

class CosineImageWindow(ImageGUI.ColorImageWindow):
    "represents a Cosine transformed image"

    def __init__(self, master, way = 1): # way: 1 = normal, 0 = inverse
        self.way = way
        masterimage = master.getnumericimage()
        for i in range(len(masterimage)):
            masterimage[i] = masterimage[i].astype(Numeric.Float)
        DCT1 = NumericImage.cosinematrix(master.height())
        DCT2 = NumericImage.cosinematrix(master.width())

        self.cosine = []
        for i in masterimage:
            if way:
                self.cosine.append(Numeric.matrixmultiply(
                    Numeric.matrixmultiply(DCT1, i._get_M()),
                    Numeric.transpose(DCT2)))
            else:
                self.cosine.append(Numeric.matrixmultiply(
                    Numeric.matrixmultiply(Numeric.transpose(DCT1),
                                           i._get_M()), DCT2))

        if way:
            titleaddon = "DCT"
        else:
            titleaddon = "inverseDCT"

        ImageGUI.ColorImageWindow.__init__(self, master,
                                           master.name + " " + titleaddon)

        if way:
            addon_structure = [
                ("separator",),
                ("command", "_Inverse Cosine Transform",
                 self.inverse_cosine, "")]
            GUI.fillmenu(self.globalmenu, addon_structure, self)

    def inverse_cosine(self, event = None):
        "Inverse Cosine Transform"
        CosineImageWindow(self, way = 0) # inverse transform

    def getrawmode(self):
        "implements for ColorImageWindow"
        if len(self.cosine) == 1: return "L"
        else: return "RGB"

    def getrawview(self):
        "implements for ColorImageWindow"
        result = []
        #maximum = max(max(max(self.cosine))) # normalization needed?
        #minimum = min(min(min(self.cosine)))
        #m = max(maximum, -minimum)
        for i in self.cosine:
            if self.way:
                #if m:
                #    result.append(NumericImage.NumImage(i * 127 / m + 128))
                #else: # avoid division by zero
                #    result.append(NumericImage.NumImage(i * 0))
                result.append(NumericImage.NumImage(i + 128))
            else:
                result.append(NumericImage.NumImage(i))
            
        if len(self.cosine) == 1: # "L"
            return result[0].toImage()
        else: # "RGB"
            return Image.merge("RGB", [result[0].toImage(),
                                       result[1].toImage(),
                                       result[2].toImage()])

    def getrawnumericimage(self):
        "implements for ColorImageWindow"
        if len(self.cosine) == 1:
            return [NumericImage.NumImage(self.cosine[0])]
        else:
            return [NumericImage.NumImage(self.cosine[0]),
                    NumericImage.NumImage(self.cosine[1]),
                    NumericImage.NumImage(self.cosine[2])]
        
    def width(self):
        "implements for ColorImageWindow"
        return self.cosine[0].shape[1]
        
    def height(self):
        "implements for ColorImageWindow"
        return self.cosine[0].shape[0]

class WalshImageWindow(ImageGUI.ColorImageWindow):
    "represents a Walsh transformed image"

    def __init__(self, master, way = 1, order = "sequential"):
        # way: 1 = normal, 0 = inverse
        # order: "sequential", "binary", "natural"
        self.way = way
        self.order = order
        masterimage = master.getnumericimage()
        for i in range(len(masterimage)):
            masterimage[i] = masterimage[i].astype(Numeric.Float)
        DWT1 = NumericImage.walshmatrix(master.height(), order)
        DWT2 = NumericImage.walshmatrix(master.width(), order)

        self.walsh = []
        for i in masterimage:
            if way:
                self.walsh.append(Numeric.matrixmultiply(
                    Numeric.matrixmultiply(DWT1, i._get_M()),
                    Numeric.transpose(DWT2)))
            else:
                self.walsh.append(Numeric.matrixmultiply(
                    Numeric.matrixmultiply(Numeric.transpose(DWT1),
                                           i._get_M()), DWT2))

        if way:
            titleaddon = "DWT"
        else:
            titleaddon = "inverseDWT"

        ImageGUI.ColorImageWindow.__init__(self, master,
                                           master.name + " " + titleaddon)

        if way:
            addon_structure = [
                ("separator",),
                ("command", "_Inverse " + order + " Walsh Transform",
                 self.inverse_walsh, "")]
            GUI.fillmenu(self.globalmenu, addon_structure, self)
            
    def inverse_walsh(self, event = None):
        "Inverse Walsh Transform"
        WalshImageWindow(self, way = 0, order = self.order)

    def getrawmode(self):
        "implements for ColorImageWindow"
        if len(self.walsh) == 1: return "L"
        else: return "RGB"

    def getrawview(self):
        "implements for ColorImageWindow"
        result = []
        maximum = max(max(max(self.walsh)))
        minimum = min(min(min(self.walsh)))
        m = max(maximum, -minimum)
        for i in self.walsh:
            if self.way and m:
                result.append(NumericImage.NumImage(i * 127 / m + 128))
            else:
                result.append(NumericImage.NumImage(i))
            
        if len(self.walsh) == 1: # "L"
            return result[0].toImage()
        else: # "RGB"
            return Image.merge("RGB", [result[0].toImage(),
                                       result[1].toImage(),
                                       result[2].toImage()])

    def getrawnumericimage(self):
        "implements for ColorImageWindow"
        if len(self.walsh) == 1:
            return [NumericImage.NumImage(self.walsh[0])]
        else:
            return [NumericImage.NumImage(self.walsh[0]),
                    NumericImage.NumImage(self.walsh[1]),
                    NumericImage.NumImage(self.walsh[2])]
        
    def width(self):
        "implements for ColorImageWindow"
        return self.walsh[0].shape[1]
        
    def height(self):
        "implements for ColorImageWindow"
        return self.walsh[0].shape[0]

class HaarImageWindow(ImageGUI.ColorImageWindow):
    "represents a Haar transformed image"

    def __init__(self, master, way = 1):
        # way: 1 = normal, 0 = inverse
        self.way = way
        masterimage = master.getnumericimage()
        for i in range(len(masterimage)):
            masterimage[i] = masterimage[i].astype(Numeric.Float)
        DHT1 = NumericImage.haarmatrix(master.height())
        DHT2 = NumericImage.haarmatrix(master.width())

        self.haar = []
        for i in masterimage:
            if way:
                self.haar.append(Numeric.matrixmultiply(
                    Numeric.matrixmultiply(DHT1, i._get_M()),
                    Numeric.transpose(DHT2)))
            else:
                self.haar.append(Numeric.matrixmultiply(
                    Numeric.matrixmultiply(Numeric.transpose(DHT1),
                                           i._get_M()), DHT2))

        if way:
            titleaddon = "DHT"
        else:
            titleaddon = "inverseDHT"

        ImageGUI.ColorImageWindow.__init__(self, master,
                                           master.name + " " + titleaddon)

        if way:
            addon_structure = [
                ("separator",),
                ("command", "_Inverse Haar Transform",
                 self.inverse_haar, "")]
            GUI.fillmenu(self.globalmenu, addon_structure, self)
            
    def inverse_haar(self, event = None):
        "Inverse Haar Transform"
        HaarImageWindow(self, way = 0)

    def getrawmode(self):
        "implements for ColorImageWindow"
        if len(self.haar) == 1: return "L"
        else: return "RGB"

    def getrawview(self):
        "implements for ColorImageWindow"
        result = []
        maximum = max(max(max(self.haar)))
        minimum = min(min(min(self.haar)))
        m = max(maximum, -minimum)
        for i in self.haar:
            if self.way and m:
                result.append(NumericImage.NumImage(i * 127 / m + 128))
            else:
                result.append(NumericImage.NumImage(i))
            
        if len(self.haar) == 1: # "L"
            return result[0].toImage()
        else: # "RGB"
            return Image.merge("RGB", [result[0].toImage(),
                                       result[1].toImage(),
                                       result[2].toImage()])

    def getrawnumericimage(self):
        "implements for ColorImageWindow"
        if len(self.haar) == 1:
            return [NumericImage.NumImage(self.haar[0])]
        else:
            return [NumericImage.NumImage(self.haar[0]),
                    NumericImage.NumImage(self.haar[1]),
                    NumericImage.NumImage(self.haar[2])]
        
    def width(self):
        "implements for ColorImageWindow"
        return self.haar[0].shape[1]
        
    def height(self):
        "implements for ColorImageWindow"
        return self.haar[0].shape[0]

