#
# Copyright 2002 Roland Stigge
#
# This file is part of ImageWorkshop
#
# ImageWorkshop is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# ImageWorkshop is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ImageWorkshop; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

# Tkinter
import Tkinter
import tkMessageBox

# Python Imaging Library (PIL)
import Image

# Numerical Python
import Numeric

# own modules
import GUI
import ImageGUI
import FilterMatrices
import NumericImage

#
# classes for local operations
#

class FilterDialog(GUI.Dialog):
    """
    asks for the convolution matrix (5 x 5)
    return values in: self.matrix (Numeric.array), optimized in size
    if successful: self.matrix != None
    """
    def body(self, master):
        matrixframe = Tkinter.Frame(master)
        matrixframe.pack(side = Tkinter.LEFT)

        Tkinter.Label(matrixframe,
                      text = "Enter convolution matrix:").grid(columnspan = 5)
        Tkinter.Label(matrixframe,
                      text = "(will be auto-normalized)").grid(row = 6,
                                                          columnspan = 5)
        self.entrymatrix = Numeric.zeros((5,5)).tolist()
        
        for i in range(5):
            for j in range(5):
                entry = Tkinter.Entry(matrixframe, width = 5)
                if i == 2 and j == 2:
                    entry.insert(Tkinter.END, "1")
                else:
                    entry.insert(Tkinter.END, "0")
        
                entry.grid(row = i + 1, column = j)

                self.entrymatrix[i][j] = entry

        self.matrix = None

        listframe = Tkinter.Frame(master) # whole left side
        listframe.pack(side = Tkinter.LEFT)

        Tkinter.Label(listframe,
                      text = "or choose from preset:").pack()
        
        listboxframe = Tkinter.Frame(listframe) # listbox with scrollbar
        listboxframe.pack()

        scrollbar = Tkinter.Scrollbar(listboxframe, orient = Tkinter.VERTICAL)
        self.listbox = Tkinter.Listbox(listboxframe,
                                       yscrollcommand = scrollbar.set)
        scrollbar.config(command = self.listbox.yview)
        scrollbar.pack(side = Tkinter.RIGHT, fill = Tkinter.Y)
        self.listbox.pack(side = Tkinter.LEFT, fill = Tkinter.BOTH, expand = 1)

        Tkinter.Button(listframe, text = "<= Apply",
                       command = self.applymatrix).pack()

        # fill listbox with names of the preset matrices
        for i in FilterMatrices.matrices:
            self.listbox.insert(Tkinter.END, i[0])
        
        return self.entrymatrix[0][0]

    def applymatrix(self):
        "callback when user clicked the Apply button to get a preset matrix"
        items = self.listbox.curselection()
        try: # workaround for bug in Python 1.5.1 which returns strings
            items = map(int, items)
        except ValueError: pass
        
        for i in range(5):
            for j in range(5):
                self.entrymatrix[i][j].delete(0, Tkinter.END)
                self.entrymatrix[i][j].insert(Tkinter.END,
                                              FilterMatrices.matrices[items[0]]
                                                [1][i][j])

    def validate(self):
        try:
            self.matrix = Numeric.zeros((5,5))
            for i in range(5):
                for j in range(5):
                    self.matrix[i][j] = float(self.entrymatrix[i][j].get())
        except ValueError:
            tkMessageBox.showwarning("Bad input", "Values must be float")
            self.matrix = None
            return 0

        # optimize matrix (size: cut "zero borders")
        while self.matrix.shape[1] > 1 and Numeric.sum(Numeric.ravel(Numeric.absolute(self.matrix[:,:1]))) == 0:
            self.matrix = self.matrix[:,1:]
        while self.matrix.shape[1] > 1 and Numeric.sum(Numeric.ravel(Numeric.absolute(self.matrix[:,-1:]))) == 0:
            self.matrix = self.matrix[:,:-1]
        while self.matrix.shape[0] > 1 and Numeric.sum(Numeric.ravel(Numeric.absolute(self.matrix[:1]))) == 0:
            self.matrix = self.matrix[1:]
        while self.matrix.shape[0] > 1 and Numeric.sum(Numeric.ravel(Numeric.absolute(self.matrix[-1:]))) == 0:
            self.matrix = self.matrix[:-1]

        return 1

class FilterImageWindow(ImageGUI.StaticImageWindow):
    "Convolution with matrix"
    def __init__(self, master, matrix):
        "input: master is a Numeric.array"
        n = master.getnumericimage()
        fheight = matrix.shape[0] # filter dimensions
        fwidth = matrix.shape[1]

        fheighth = fheight / 2 # speed optimization in inner loop
        fwidthh = fwidth / 2
        
        # analyse matrix
        plussum = 0.0    # absolute values
        minussum = 0.0
        for i in range(fheight):
            for j in range(fwidth):
                if matrix[i, j] > 0:
                    plussum += matrix[i, j]
                if matrix[i, j] < 0:
                    minussum -= matrix[i, j]
        
        result = []
        # convolve image
        for channel in range(len(n)): # for each channel
            newchannel = Numeric.zeros((master.height(), master.width()))
            for r in range(master.height()):
                for c in range(master.width()):
                    sum = 0.0
                    for i in range(fheight):
                        for j in range(fwidth):
                            try:
                                sum += n[channel].M[r - i + fheighth,
                                                    c - j + fwidthh] * matrix[i, j]
                            except IndexError:
                                pass
                    newchannel[r, c] = sum
            # normalize channel
            newchannel = newchannel / max(plussum, minussum)
            if minussum: # -> offset
                newchannel = newchannel / 2 + 128
            
            result.append(newchannel)

        if len(result) == 1: # "L"
            self.image = NumericImage.NumImage(result[0]).toImage()
        else: # "RGB"
            self.image = Image.merge("RGB",
                                [NumericImage.NumImage(result[0]).toImage(),
                                 NumericImage.NumImage(result[1]).toImage(),
                                 NumericImage.NumImage(result[2]).toImage()])

        ImageGUI.StaticImageWindow.__init__(self, master,
                                            master.name + " [filtered]")
        
class RankDialog(GUI.Dialog):
    """
    asks for the rank order mode:
    0 = median, 1 = erosion, 2 = dilatation, 3 = opening, 4 = closing filter
    return values in: self.mode, self.size
    if successful: self.mode != None
    """
    def body(self, master):
        self.modevar = Tkinter.IntVar()
        self.modevar.set(0)
        first = Tkinter.Radiobutton(master, text = "Median", value = 0,
                                    variable = self.modevar).grid(row = 0,
                                          sticky = Tkinter.W, columnspan = 2)
        Tkinter.Radiobutton(master, text = "Erosion", value = 1,
                            variable = self.modevar).grid(row = 1,
                                          sticky = Tkinter.W, columnspan = 2)
        Tkinter.Radiobutton(master, text = "Dilatation", value = 2,
                            variable = self.modevar).grid(row = 2,
                                          sticky = Tkinter.W, columnspan = 2)
        Tkinter.Radiobutton(master, text = "Opening", value = 3,
                            variable = self.modevar).grid(row = 3,
                                          sticky = Tkinter.W, columnspan = 2)
        Tkinter.Radiobutton(master, text = "Closing", value = 4,
                            variable = self.modevar).grid(row = 4,
                                          sticky = Tkinter.W, columnspan = 2)
        Tkinter.Label(master, text = "Window size:").grid(row = 5)
        self.sizeentry = Tkinter.Entry(master)
        self.sizeentry.insert(Tkinter.END, "3")
        self.sizeentry.grid(row = 5, column = 1)
        
        self.mode = None
        
        return first

    def validate(self):
        self.mode = self.modevar.get()
        try:
            self.size = int(self.sizeentry.get())
        except ValueError:
            tkMessageBox.showwarning("Bad input", "Values must be float")
            self.mode = None
            return 0

        return 1

class RankImageWindow(ImageGUI.StaticImageWindow):
    def __init__(self, master, mode, size):
        modes = ["Median", "Erosion", "Dilatation", "Opening", "Closing"]
        n = master.getnumericimage()

        listsize = size ** 2
        
        result = []
        for channel in range(len(n)): # for each channel
            newchannel = Numeric.zeros((master.height(), master.width()))
            for r in range(master.height()):
                for c in range(master.width()):
                    list = []
                    for i in range(size):
                        for j in range(size):
                            try:
                                list.append(n[channel].M[r + i - size / 2,
                                                         c + j - size / 2])
                            except IndexError:
                                list.append(0)
                    list.sort()
                    if mode == 0: # Median
                        newchannel[r, c] = list[size / 2]
                    elif mode in [1, 3]: # Erosion (and Opening)
                        newchannel[r, c] = list[0]
                    elif mode in [2, 4]: # Dilatation (and Closing)
                        newchannel[r, c] = list[-1]
                        
            if mode in [3, 4]: # Dilatation for Opening, Erosion for Closing
                oldchannel = newchannel
                newchannel = Numeric.zeros((master.height(), master.width()))
                for r in range(master.height()):
                    for c in range(master.width()):
                        list = []
                        for i in range(size):
                            for j in range(size):
                                try:
                                    list.append(oldchannel[r + i - size / 2,
                                                           c + j - size / 2])
                                except IndexError:
                                    list.append(0)
                        list.sort()
                        if mode == 3: # Dilatation for Opening
                            newchannel[r, c] = list[-1]
                        else: # Erosion for Closing
                            newchannel[r, c] = list[0]
                        
            result.append(newchannel)

        if len(result) == 1: # "L"
            self.image = NumericImage.NumImage(result[0]).toImage()
        else: # "RGB"
            self.image = Image.merge("RGB",
                                [NumericImage.NumImage(result[0]).toImage(),
                                 NumericImage.NumImage(result[1]).toImage(),
                                 NumericImage.NumImage(result[2]).toImage()])

        ImageGUI.StaticImageWindow.__init__(self, master,
                                            master.name + " [" +
                                            modes[mode] + "]")
        
